# pyuvm uses the Python logging system to do reporting.
# Still, we need this base class to be true to the hierarchy.
# Every instance of a child class has its own logger.
#
# There may be a need to implement uvm_info, uvm_error,
# uvm_warning, and uvm_fatal, but it would be best to
# first see how the native Python logging system does the job.

from pyuvm.s05_base_classes import uvm_object
import logging


# 6.2.1
class uvm_report_object(uvm_object):
    """ The basis of all classes that can report """

    def __init__(self, name):
        super().__init__(name)
        uvm_root_logger = logging.getLogger('uvm')
        # Every object gets its own logger
        self.logger = uvm_root_logger.getChild(self.get_full_name())
        self.logger.setLevel(level=logging.INFO)  # Default is to print INFO
        # We are not sending log messages up the hierarchy
        self.logger.propagate = False
        handler = logging.StreamHandler()
        # Don't let the handler interfere with logger level
        handler.setLevel(logging.NOTSET)
        # Make log messages look like UVM messages
        formatter = logging.Formatter(
            "%(levelname)s: %(filename)s(%(lineno)d)[%(name)s]: %(message)s")
        handler.setFormatter(formatter)
        self.add_logging_handler(handler)
        pass

    def set_logging_level(self, logging_level):
        """ Sets the logger level """
        self.logger.setLevel(logging_level)

    def add_logging_handler(self, handler):
        """ Adds a handler """
        assert isinstance(handler, logging.Handler), \
            f"You must pass a logging.Handler not {type(handler)}"
        self.logger.addHandler(handler)

    def remove_logging_handler(self, handler):
        """ Removes a specific handler  """
        assert isinstance(handler, logging.Handler), \
            f"You must pass a logging.Handler not {type(handler)}"
        self.logger.removeHandler(handler)

    def set_formatter_on_handlers(self, formatter):
        """ Set an identical formatter on all handlers """
        assert isinstance(formatter, logging.Formatter), \
            f"You must pass a logging.Formatter not {type(formatter)}"
        for handler in self.logger.handlers:
            handler.setFormatter(formatter)



"""
This file defines the UVM base classes
"""
import sys

try:
    import copy
    import pyuvm.error_classes as error_classes
    import pyuvm.utility_classes as utility_classes
    from pyuvm.s08_factory_classes import uvm_factory
except ModuleNotFoundError as mnf:
    print(mnf)
    sys.exit(1)


# 5.3.1
class uvm_object(utility_classes.uvm_void):
    """ The most basic UVM object """

    # 5.3.2
    def __init__(self, name=''):
        assert (isinstance(name, str)), \
            f"{name} is not a string it is a {type(name)}"
        self.set_name(name)

    # 5.3.3.1
    def get_uvm_seeding(self):
        """Not implemented"""
        raise error_classes.UVMNotImplemented(
            'get_uvm_seeding not implemented')

    # 5.3.3.2
    def set_uvm_seeding(self, enable):
        """ Not implemented """
        raise error_classes.UVMNotImplemented(
            'set_uvm_seeding not implemented')

    # 5.3.3.3
    def reseed(self):
        """ Not implemented """
        raise error_classes.UVMNotImplemented('reseed not implemented')

    # 5.3.3.4
    def get_name(self):
        """
        Return the name of this object as passed by the constructor
        """
        assert (self._obj_name is not None), \
            f"Internal error. {str(self)} has no name"
        return self._obj_name

    # 5.3.4.1
    def set_name(self, name):
        """
        Set the name
        """
        assert (isinstance(name, str)), "Must set the name to a string"
        self._obj_name = name

    # 5.3.4.3
    def get_full_name(self):
        """
        The full name for a uvm_object is simply the name
        """
        return self.get_name()

    # 5.3.4.4
    def get_inst_id(self):
        """
        Returns the python ID which fits the bill
        for what the ID is supposed to be.
        """
        return id(self)

    # 5.3.4.5
    def get_type(self):
        """
        Not implemented because Python can implement the factory without
        these shenanigans.
        """
        raise error_classes.UVMNotImplemented(
            'Python provides better ways to do this '
            'so the uvm_object_wrapper is unimplemented')

    # 5.3.4.6
    def get_object_type(self):
        """
        Not implemented because Python can implement the factory without
        these shenanigans.
        """
        raise error_classes.UVMNotImplemented(
            'Python provides better ways to do this '
            'so the uvm_object_wrapper is unimplemented')

    # 5.3.4.7
    def get_type_name(self):
        """
        Returns types  __name__ magic variable
        """
        return type(self).__name__

    # 5.3.5.1
    @classmethod
    def create(cls, name):
        """
        :return: new object from factory
        """
        new_obj = uvm_factory().create_object_by_type(cls, name=name)
        return new_obj

    # 5.3.5.2
    def clone(self):
        """
        Unlike the clone in the SystemVerilog UVM, this clone()
        uses copy.deepcopy so the user does not need to override
        do_copy()
        """
        return copy.deepcopy(self)

    # 5.3.6.1
    def print(self):
        """
        Not implemented. Use __str__() and print()
        """
        raise error_classes.UVMNotImplemented(
            'There are better ways to do printing in Python')

    # 5.3.6.2
    def sprint(self):
        """ Not implemented. use __str__() and print()"""
        raise error_classes.UVMNotImplemented(
            'There are better ways to do printing in Python')

    # 5.3.6.3
    def do_print(self):
        """ not implemented. Use __str__() and print()"""
        raise error_classes.UVMNotImplemented(
            'There are better ways to do printing in Python')

    # 5.3.6.4
    def convert2string(self):
        """
        Returns the result of self.__str__().
        Perhaps better to just use __str__() directly?
        :return __str__()
        """
        return self.__str__()

    # 5.3.7
    def record(self):
        """
        Not implemented as we are not in a simulator.
        """
        raise error_classes.UVMNotImplemented(
            'Python does not run in the simulator, so no recording')

    # 5.3.7.2
    def do_record(self):
        """
        Not implemented as we are not in a simulator
        """
        raise error_classes.UVMNotImplemented(
            'Python does not run in the simulator, so no recording')

    # 5.3.8.1
    def copy(self, rhs):
        """
        Returns do_copy() though that just calls copy.deepcopy()
        """
        return self.do_copy(rhs)

    # 5.3.8.2
    def do_copy(self, rhs):
        """
        Must be overridden, though copy.deepcopy() does the equivalent
        and is a more Pythonic solution
        """
        raise error_classes.UsePythonMethod('Use the copy.deepcopy() '
                                            'method to implement this to '
                                            'support the copy module or '
                                            'override this method.')

    # 5.3.9.1
    def compare(self, rhs):
        """

        """
        return self.do_compare(rhs)

    # 5.3.9.2
    def do_compare(self, rhs):
        """
        Recommend overriding __eq__() rather than this method.
        """
        return self.__eq__(rhs)

    # 5.3.10.1
    def pack(self):
        """
        Not implemented yet. There are Pythonic solutions to this.
        """
        raise error_classes.UsePythonMethod(
            "use struct, pickle, json, or yaml.")

    # 5.3.10.1
    def pack_bytes(self):
        """
        Not implemented yet. There are Pythonic solutions to this.
        """
        raise error_classes.UsePythonMethod(
            "use struct, pickle, json, or yaml.")

    # 5.3.10.1
    def pack_ints(self):
        """
        Not implemented yet. There are Pythonic solutions to this.
        """
        raise error_classes.UsePythonMethod(
            "use struct, pickle, json, or yaml.")

    # 5.3.10.1
    def pack_longints(self):
        """
        Not implemented yet. There are Pythonic solutions to this.
        """
        raise error_classes.UsePythonMethod(
            "use struct, pickle, json, or yaml.")

    # 5.3.10.2
    def do_pack(self):
        """
        Not implemented yet. There are Pythonic solutions to this.
        """
        raise error_classes.UsePythonMethod(
            "use struct, pickle, json, or yaml.")

    # 5.3.11.1
    def unpack(self):
        """
        Not implemented yet. There are Pythonic solutions to this.
        """
        raise error_classes.UsePythonMethod(
            "use pickle, json, or yaml.")

    # 5.3.14.1
    def push_active_policy(self):
        """
        Not implemented yet.
        """
        raise error_classes.UVMNotImplemented(
            "policies not implemented yet")

    # 5.3.14.2
    def pop_active_policy(self):
        """
        Not implemented yet.
        """
        raise error_classes.UVMNotImplemented("policies not implemented yet")

    # 5.3.14.3
    def get_active_policy(self):
        """
        Not implemented yet.
        """
        raise error_classes.UVMNotImplemented("policies not implemented yet")

    # 5.3.11.1
    def unpack_bytes(self):
        """
        Not implemented yet. There are Pythonic solutions to this.
        """
        raise error_classes.UsePythonMethod(
            "use struct, pickle, json, or yaml.")

    # 5.3.11.1
    def unpack_ints(self):
        """
        Not implemented yet. There are Pythonic solutions to this.
        """
        raise error_classes.UsePythonMethod(
            "use struct, pickle, json, or yaml.")

    # 5.3.11.1
    def unpack_longints(self):
        """
        Not implemented yet. There are Pythonic solutions to this.
        """
        raise error_classes.UsePythonMethod(
            "use struct, pickle, json, or yaml.")

    # 5.3.11.2
    def do_unpack(self):
        """
        Not implemented yet. There are Pythonic solutions to this.
        """
        raise error_classes.UsePythonMethod(
            "use struct, pickle, json, or yaml.")

    # 5.3.12
    def set_local(self):
        """
        Not implemented use Python getattr and setattr.
        """
        raise error_classes.UsePythonMethod(
            'The getattr and setattr functions handle this')

    # 5.3.13.1
    def do_execute_op(self, op):
        """
        Not implemented.
        """
        raise error_classes.UsePythonMethod(
            'Not needed in Python')


# 5.3.13
class uvm_field_op:
    """
    We do not implement the UVM field op as this is a
    UVM way of providing field-based functionality
    that can better be implemented using Python functionality.
    """

    def __new__(cls, *args, **kwargs):
        raise error_classes.UsePythonMethod(
            'Python has simpler ways of handling field function.')


# 5.3.14
class uvm_policy:
    """
    The uvm_policy is used to add functionality to
    SystemVerilog that already exists
    in Python. It is not needed in pyuvm.
    """

    def __new__(cls, *args, **kwargs):
        raise error_classes.UsePythonMethod(
            'Python has simpler ways of handling '
            'functionality provided by policies.')


# 5.4.1
class uvm_transaction(uvm_object):
    """
    Transactions without interface to logging or waveforms.
    """

    # 5.4.2.1
    def __init__(self, name="", initiator=None):
        """
        :param name: Object name
        :param initiator: component that is initiator
        """
        super().__init__(name)
        self.set_initiator(initiator)
        self.transaction_id = id(self)

    def set_id_info(self, other):
        """
        Set transaction_id from other
        :param other: uvm_transaction
        :return: None
        """
        self.transaction_id = other.transaction_id

    def set_initiator(self, initiator):
        """
        5.4.2.14
        :param initiator: initiator to set
        :return: None
        """
        self._initiator = initiator

    def get_initiator(self):
        """
        5.4.2.15
        :return: initiator
        """
        return self._initiator

    def __not_implemented(self):
        raise error_classes.UVMNotImplemented(
            'This method is not implemented at this time.')

    # 5.4.2.2
    def accept_tr(self, time):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.3
    def do_accept_tr(self):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.5
    def begin_tr(self, begin_time=0, parent_handle=None):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.5
    def do_begin_tr(self):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.6
    def end_tr(self, end_time=0, free_handle=True):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.7
    def do_end_tr(self):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.8
    def get_tr_handle(self):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.9
    def enable_recording(self):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.10
    def disable_recording(self):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.11
    def is_recording_enabled(self):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.12
    def is_active(self):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.13
    def get_event_pool(self):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.16
    def get_accept_time(self):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.16
    def get_begin_time(self):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.16
    def get_end_time(self):
        """
        Not implemented
        """
        self.__not_implemented()

    # 5.4.2.17
    def set_transaction_id(self, txn_id):
        """
        Sets  variable transaction_id
        """
        assert (isinstance(txn_id, int)), "Transaction ID must be an integer."
        self.transaction_id = txn_id

    # 5.4.2.18
    def get_transaction_id(self):
        """
        Returns  variable transaction_id
        """
        if self.transaction_id is None:
            return id(self)
        else:
            return self.transaction_id




class UVMError(Exception):
    """
    All UVM Errors
    """


class UVMNotImplemented(UVMError):
    """For methods that we haven't yet implemented."""


class UsePythonMethod(UVMError):
    """
    For cases where the user should use a Python
    method rather than a UVM method.  For example
    use __str__() instead of convert2string()
    """


class UVMFactoryError(UVMError):
    """ For cases where a type is not registered with the factory"""


class UVMTLMConnectionError(UVMError):
    """For problems connecting TLM"""


class UVMBadPhase(UVMError):
    """Errors in phasing"""


class UVMSequenceError(UVMError):
    """Errors using sequences"""


class UVMConfigError(UVMError):
    """Errors using the config_db"""


class UVMConfigItemNotFound(UVMError):
    """Couldn't find something in config_db"""


class UVMFatalError(UVMError):
    """ Used to dump out of the testbench"""



import pyuvm.utility_classes as utility_classes
import pyuvm.error_classes as error_classes
import logging
import fnmatch


# pyuvm refactors the factory, taking advantage Python's
# superiority in terms of OOP features and lack of types.

# Implementing section 8 in the IEEE Specification
# The IEEE spec assumes that UVM is being written in SystemVerilog,
# a language that does not allow you to control how classes get defined.
# Python has more control in this area. With Python we can set things up
# so that any class that extends uvm_void automatically gets registered
# into the factory.

# Therefore there is no need for 8.2.2 the type_id, that
# is a SystemVerilog artifact as is 8.2.3
# There is also no need for 8.2.4, the uvm_object_registry.
# The FactoryMeta class causes all classes from uvm_void
# down to automatically register themselves with the factory
# by copying themselves into a dict.

# However there is a need to provide the methods in 8.


# 8.3.1.1
class uvm_factory(metaclass=utility_classes.Singleton):
    """
    The uvm_factory is a singleton that delivers all factory functions.
    """

    # 8.3.1.2.1 get
    # There is no get() method in Python singletons.  Instead you instantiate
    # the singleton as normal and you automatically get the singleton.

    # 8.3.1.3 register
    # Not implemented
    # There is no register in pyuvm.  Instead the factory builds its class
    # database through introspection.

    def __init__(self):
        self.fd = utility_classes.FactoryData()
        self.logger = logging.getLogger("Factory")
        self.all_types = 1

    def __set_override(self, original, override, path=None):
        if original not in self.fd.overrides:
            self.fd.overrides[original] = utility_classes.Override()
        self.fd.overrides[original].add(override, path)

    # 8.3.1.3
    def set_inst_override_by_type(self, original_type, override_type,
                                  full_inst_path):
        """
        Override an instance with a new type if original type is at that path
        :param original_type: The original type being overridden
        :param override_type: The overriding type
        :param full_inst_path: The inst where this happens.
        """

        # The intention here is to only override when a type of original_type
        # is at the full_inst_path provided. If someone stores a different
        # class at full_inst_path then the override will not happen.
        #
        #
        # We capture this by storing the original and override types as a
        # tuple at the full inst path.  Later we'll retrieve the tuple
        # and check the type of the object at the full_inst_path.
        #
        # instance_overrides is an OrderedDict, so we will check
        # the paths in the order they are registered later.

        assert issubclass(original_type, utility_classes.uvm_void),\
            "You tried to override a non-uvm_void class"
        assert issubclass(override_type,
                          utility_classes.uvm_void),\
            "You tried to use a non-uvm_void class as an override"
        self.__set_override(original_type, override_type, full_inst_path)

    # 8.3.1.4.1
    def set_inst_override_by_name(self, original_type_name,
                                  override_type_name, full_inst_path):
        """
        Override a specific instance  using strings that contain
        the names of the types.
        :param original_type_name: the name of type being replaced
        :param override_type_name: the name of the substitute type
        :param full_inst_path: The path to the instance
        :return:
        """

        # Here we use the names of classes instead of the classes.
        # The original_name doesn't need to be the name of a class,
        # it can be an arbitrary string. The override_name must
        # be the name of a class.
        #
        # Later we will retrieve this by searching through the
        # keys for a match to a path and then checking that
        # the name given in the search matches the original_name

        assert isinstance(full_inst_path, str), \
            "The inst_path must be a string"
        assert isinstance(original_type_name, str), \
            "Original_name must be a string"
        assert isinstance(override_type_name, str), \
            "Override_name must be a string"
        try:
            override_type = self.fd.classes[override_type_name]
        except KeyError:
            raise error_classes.UVMFactoryError(
                f"{override_type_name}" + " has not been defined.")

        # Set type override by name can use an arbitrary string as a key
        # instead of a type. Fortunately Python dicts don't care about
        # the type of the key.
        try:
            original_type = self.fd.classes[original_type_name]
        except KeyError:
            original_type = original_type_name

        self.__set_override(original_type, override_type, full_inst_path)

    # 8.3.1.4.2
    def set_type_override_by_type(self, original_type, override_type,
                                  replace=True):
        """
        Override one type with another type globally
        :param original_type: The original type to be overridden
        :param override_type: The new type that will override it
        :param replace: If the override exists, only replace it if this is True
        """
        assert issubclass(original_type, utility_classes.uvm_void),\
            "You tried to override a non-uvm_void class"
        assert issubclass(override_type,
                          utility_classes.uvm_void),\
            "You tried to use a non-uvm_void class as an override"
        if (original_type not in self.fd.overrides) or replace:
            self.__set_override(original_type, override_type)

    # 8.3.1.4.2
    def set_type_override_by_name(self, original_type_name,
                                  override_type_name, replace=True):
        """
        Override one type with another type globally using strings
        containing the type names.

        :param original_type_name: The name of the type to be
        overridden or an arbitrary string.
        :param override_type_name: The name of the overriding type.
        It must have been declared.
        :param replace: If the override already exists only
        replace if this is True
        :return:
        """
        assert isinstance(original_type_name, str),\
            "Original_name must be a string"
        assert isinstance(override_type_name, str),\
            "Override_name must be a string"
        try:
            override_type = self.fd.classes[override_type_name]
        except KeyError:
            raise error_classes.UVMFactoryError(
                f"{override_type_name}" + " has not been defined.")

        # Set type override by name can use an arbitrary string as a key
        # instead of a type

        # Fortunately Python dicts don't care about the type of the key.
        try:
            original_type = self.fd.classes[original_type_name]
        except KeyError:
            original_type = original_type_name

        if (original_type not in self.fd.overrides) or replace:
            self.__set_override(original_type, override_type)

    def __find_override(self, requested_type, parent_inst_path="", name=""):
        """
        An internal function that finds overrides

        :param requested_type: The type that could be overridden
        :param parent_inst_path: The parent inst path for an override
        :param name: The name of the object, concatenated with parent
        inst path for override
        :return: either the requested_type or its override
        """
        if not isinstance(requested_type, str):
            assert (issubclass(requested_type, utility_classes.uvm_void)), \
                f"You must create uvm_void descendants not {requested_type}"

        if parent_inst_path == "":
            inst_path = name
        elif name != "":
            inst_path = parent_inst_path + "." + name
        else:
            inst_path = parent_inst_path

        new_cls = self.fd.find_override(requested_type, inst_path)
        if isinstance(new_cls, str):
            self.logger.error(
                f'"{new_cls}" is not declared and is not an override string')
            return None
        else:
            return new_cls

    def create_object_by_type(self, requested_type, parent_inst_path="",
                              name=""):
        """
        8.3.1.5 Creation
        :param requested_type: The type that we request but that can be
        overridden
        :param parent_inst_path: The get_full_name path of the parent
        :param name: The name of the instance requested_type("name")
        :return: Type that is child of uvm_object.
        Python does not create zero-length strings as defaults.
        It puts the None object there. That's
        what we're going to do.
        """
        new_type = self.__find_override(requested_type, parent_inst_path, name)
        if new_type is None:
            return None
        return new_type(name)

    # 8.3.1.5
    def create_object_by_name(self, requested_type_name,
                              parent_inst_path="", name=""):
        """
        Create an object using a string to define its uvm_object type.

        :param requested_type_name: the type that could be overridden
        :param parent_inst_path: A path if we are checking for inst overrides
        :param name: The name of the new object.
        :return: A uvm_object with the name given
        """
        try:
            requested_type = utility_classes.FactoryData().classes[requested_type_name] # noqa
        except KeyError:
            requested_type = requested_type_name

        new_obj = self.create_object_by_type(requested_type, parent_inst_path,
                                             name)
        return new_obj

    # 8.3.1.5
    def create_component_by_type(self, requested_type, parent_inst_path="",
                                 name="", parent=None):
        """
        Create a component of the requested uvm_component type.

        :param requested_type: Type type to be overridden
        :param parent_inst_path: The inst path if we are looking
        for inst overrides
        :param name: Concatenated with parent_inst_path if it
        exists for inst overrides
        :param parent: The parent component
        :return: a uvm_component with the name an parent given.
        """

        if name is None:
            raise error_classes.UVMFactoryError(
                "Parameter name must be specified in function call.")

        new_type = self.__find_override(requested_type, parent_inst_path, name)

        if new_type is None:
            return None

        new_comp = new_type(name, parent)
        return new_comp

    # 8.3.1.5
    def create_component_by_name(self, requested_type_name,
                                 parent_inst_path="", name="", parent=None):
        """
        Create a components using the name of the requested uvm_component type

        :param requested_type_name: the type that could be overridden
        :param parent_inst_path: A path if we are checking for inst overrides
        :param name: The name of the new object.
        :param parent: The component's parent component
        :return: A uvm_object with the name given
        """
        if name is None:
            raise error_classes.UVMFactoryError(
                "Parameter name must be specified in create_component_by_name")

        try:
            requested_type = utility_classes.FactoryData().classes[requested_type_name]  # noqa
        except KeyError:
            requested_type = requested_type_name

        new_obj = self.create_component_by_type(requested_type,
                                                parent_inst_path,
                                                name, parent)
        return new_obj

    # 8.3.1.6.1
    def set_type_alias(self, alias_type_name, original_type):
        """
        Not implemented as it does not seem to exist in SystemVerilog UVM

        :param alias_type_name:A string that will reference the original type
        :param original_type:The original type toe be referenced
        :return:None
        """
        # This method does not seem to be implemented in SystemVerilog
        # so I'm skipping it now.
        raise error_classes.UVMNotImplemented(
            "set_type_alias is not implemented in SystemVerilog")

    # 8.3.1.6.2
    def set_inst_alias(self, alias_type_name, original_type, full_inst_path):
        """
        Not implemented as it does not seem to exist in SystemVerilog UVM

        :param alias_type_name:A string that will reference the original type
        :param original_type:The original type toe be referenced
        :param full_inst_path: The instance path where this alias is active
        :return:None
        """
        # This method does not seem to be implemented in SystemVerilog
        # so I'm skipping it now.
        raise error_classes.UVMNotImplemented(
            "set_type_inst is not implemented in SystemVerilog")

    # 8.3.1.7 Introspection

    # 8.3.1.7.1
    def find_override_by_type(self, requested_type, full_inst_path):
        """
        Given a type and instance path, return the override class object.

        :param requested_type: The type whose override you want
        :param full_inst_path: The inst path where one looks
        :return: class object
        """
        override = self.__find_override(requested_type, full_inst_path)
        return override

    # 8.3.1.7.1
    def find_override_by_name(self, requested_type_name, full_inst_path):
        """
        Given a path and the name of a class return its overriding class object

        :param requested_type_name:
        :param full_inst_path:
        :return: class object
        """
        assert (isinstance(requested_type_name, str)), \
            f"requested_type_name must be a string not a {type(requested_type_name)}"  # noqa
        requested_type = None  # Appeasing the linter
        try:
            requested_type = self.fd.classes[requested_type_name]
        except KeyError:
            error_classes.UVMFactoryError(
                f"{requested_type_name} is not a defined class name")
        return self.find_override_by_type(requested_type, full_inst_path)

    # 8.3.1.7.2
    def find_wrapper_by_name(self):
        """
        There are no wrappers in pyuvm so this is not implemented.
        """
        raise error_classes.UVMNotImplemented(
            "There are no wrappers in pyuvm. "
            "So find_wrapper_by_name is not implemented")

    # 8.3.1.7.3
    def is_type_name_registered(self, type_name):
        """
        Checks that a type of this name is registered with the factory.

        :param type_name: string that is name of a type
        :return: boolean
        """
        assert (isinstance(type_name,
                           str)), "is_type_name_registered() takes a"
        " string as its argument not {type(type_name)}"
        return type_name in self.fd.classes

    # 8.3.1.7.4
    def is_type_registered(self, uvm_type):
        """
        Checks that a type is registered. The argument is named "obj" in
        the spec, but that name is ridiculous and confusing.

        :param uvm_type: The type to be checked
        :return: boolean
        """
        assert (issubclass(uvm_type, utility_classes.uvm_void)), \
            "is_type_registered() takes a subclass of uvm_void "
        "as its argument not {type(uvm_type)}"
        return uvm_type in self.fd.classes.values()

    def __str__(self):
        """
        Returns the Pythonic string
        Set uvm_factory().all_types to a value to control the string.
        The default is 1

        uvm_factory().all_types = 0 : overrides
        uvm_factory().all_types = 1 : user defined types + above
        uvm_factory().all_types = 2 : uvm_* types + above

        :return: String containing factory data
        """
        assert (0 <= self.all_types <= 2), \
            "uvm_factory().all_type must be 0, 1, or 2"
        factory_str = "--- overrides "
        if self.all_types != 0:
            factory_str += "+ user defined types "
        if self.all_types == 2:
            factory_str += "+ uvm_* types "
        factory_str += "---\n\n"

        if len(self.fd.overrides) > 0:
            factory_str += "Overrides:\n"
            for inst in self.fd.overrides:
                factory_str += f"{inst.__name__:25}" + ": " + str(self.fd.overrides[inst])  # noqa
                factory_str += "\n"
        # Need to add 1 and 2
        user_list = [self.fd.classes[cls].__name__
                     for cls in self.fd.classes
                     if not fnmatch.fnmatch(self.fd.classes[cls].__name__, "uvm_*")]  # noqa
        uvm_list = [self.fd.classes[cls].__name__
                    for cls in self.fd.classes
                    if fnmatch.fnmatch(self.fd.classes[cls].__name__, "uvm_*")]
        if self.all_types > 0:
            factory_str += "\n" + "-" * 25 + "\nUser Defined Types:\n"
            factory_str += "\n".join(user_list)

        if self.all_types == 2:
            factory_str += "\n" + "-" * 25 + "\nUVM Types:\n"
            factory_str += "\n".join(uvm_list)

        return factory_str

    # 8.3.1.7.5
    def print(self, all_types=1):
        """
        Prints the factory data using all_types to
        control the amount of output. The uvm_factory().all_types
        variable can control this for __str__()

        all_types = 0 : overrides
        all_types = 1 : user defined types + above
        all_types = 2 : uvm_* types + above

        :return: None
        """
        saved_all_types = self.all_types  # Avoiding a side effect
        self.all_types = all_types
        print(self)
        self.all_types = saved_all_types

# 8.3.1.8 Usage
# All the elements of usage have been implemented in pyuvm
# The biggest difference is that all uvm_void classes get registered
# with the factory automatically.

# 8.3.2
# Not implemented.  There is no need for a uvm_object_wrapper
# in Python. The user will never notice
# the difference



from pyuvm.s06_reporting_classes import uvm_report_object
from pyuvm.s08_factory_classes import uvm_factory
from pyuvm.s09_phasing import uvm_common_phases, uvm_run_phase, uvm_build_phase
from pyuvm import error_classes
from pyuvm import utility_classes
import logging
import fnmatch
import string


class uvm_component(uvm_report_object):
    """13.1.1 Class Declaration

The specification calls for uvm_component to extend
uvm_report_object. However, pyuvm uses the logging
module orthogonally to class structure. It may be
that in future we find a reason to wrap the basic
logging package in the uvm code, but at this point we are
better off leaving logging to itself.

The choice then becomes whether to create a
uvm_report_object class as a placeholder to preserve
the UVM reference manual hierarchy or to code what
is really going on.  We've opted for the latter.

    """
    component_dict = {}

    @classmethod
    def clear_components(cls):
        cls.component_dict = {}

    def __init__(self, name, parent):
        """
        13.1.2.1---This is new() in the IEEE-UVM, but we mean
        the same thing with __init__()
        """

        self.__children = {}
        if parent is None and name != 'uvm_root':
            parent = uvm_root()
        self.parent = parent
        if parent is not None:
            parent.add_child(name, self)
        self.print_enabled = True  # 13.1.2.2
        super().__init__(name)

        # Cache the hierarchy for easy access
        if name != 'uvm_root':
            uvm_component.component_dict[self.get_full_name()] = self

    def clear_children(self):
        self.__children = {}

    def clear_hierarchy(self):
        self._parent = None
        self.clear_children()

    def do_execute_op(self, op):
        raise error_classes.UVMNotImplemented("Policies not implemented")

    @classmethod
    def create(cls, name="", parent=None):
        if parent is None:
            parent_inst_path = ""
        else:
            parent_inst_path = parent.get_full_name()

        new_comp = uvm_factory().create_component_by_type(
            cls, parent_inst_path=parent_inst_path,
            name=name, parent=parent)
        return new_comp

    def get_parent(self):
        """
        :return: parent object
        13.1.3.1--- No 'get_' prefix
        """
        return self._parent

    def raise_objection(self):
        utility_classes.ObjectionHandler().raise_objection(self)

    def drop_objection(self):
        utility_classes.ObjectionHandler().drop_objection(self)

    def cdb_set(self, label, value, inst_path="*"):
        """
        Store an object in the config_db.

        :param value: The object to store
        :param label: The label to use to retrieve it
        :param inst_path: A path with globs or if left blank
                          the get_full_name() path
        """

        ConfigDB().set(self, inst_path, label, value)

    def cdb_get(self, label, inst_path=""):
        """
        Retrieve an object from the config_db using this components
        get_full_name() path. Can find objects stored with wildcards

        :param inst_path: The path below this component
        :param label: The label used to store the value
        :return: The object at this path stored at the label
        """
        datum = ConfigDB().get(self, inst_path, label)
        return datum

    @property
    def parent(self):
        return self.get_parent()

    @parent.setter
    def parent(self, parent):
        if parent is not None:
            assert (isinstance(parent, uvm_component)), \
                f" {parent} is of type {type(parent)}"
        assert (parent != self), \
            f'Cannot make a {self.get_name()} its own parent.  That is incest.'
        self._parent = parent

    def get_full_name(self):
        """
        :return: Name concatenated to parent name.
        13.1.3.2
        """
        if self.get_name() is None or self.get_name() == 'uvm_root':
            return ''
        if self._parent is None:
            fullname = ""
        else:
            fullname = self._parent.get_full_name()
        if len(fullname) == 0:
            fullname = self.get_name()
        else:
            fullname = fullname + "." + self.get_name()
        return fullname

    # Children in pyuvm
    # 13.1.3.4 modified to be pythonic
    """
    UVM components contain a list of child components as
    the basis for creating the UVM hierarchy.  However
    the UVM accesses those children in an unPythonic
    way.  Instead of asking for an iterator of children,
    which is what Python does and what we'll do here, the
    UVM uses `get_child()`, `get_first_child()`, and
    `get_next_child()` to create iterative-like behavior,
    but there is no sense in recreating this if we're
    in a language like Python with powerful iterative
    functionality.

    We will store children in a Python dict so as to
    implement get_child(), but we will not implement
    get_first_child() or get_next_child().  We will
    implement __iter__ so that the user can use
    the component as an iterator.
    """

    def get_children(self):
        """
        13.1.3.3
        :return: dict with children
        """
        return list(self.children)

    def add_child(self, name, child):
        assert (name not in self.__children), \
            f"{self.get_full_name()} already has a child named {name}"
        self.__children[name] = child
        pass

    @property
    def hierarchy(self):
        """
        We return a generator to find the children.
        This is more pythonic and saves memory for large hierarchies.
        :return: An ordered list of components top to bottom.
        """
        yield self
        for child in self.children:
            assert isinstance(child, uvm_component)
            yield child
            for grandchild in child.children:
                assert isinstance(grandchild, uvm_component, )
                yield grandchild

        # The UVM relies upon a hokey iteration system to get the children
        # out of a component class. You get the name of the first child and
        # then pass it to get_next_child to get the name of the next
        # child. This continues until get_next_child returns zero.
        #
        # Python has a rich iteration system and it would be foolish to
        # eschew it. So we are not going to implement the above.  Instead
        # the children() method is a generator that allows you to loop
        # through the children.

    @property
    def children(self):
        """
        13.1.3.4
        Implements the intention of this requirement
        without the approach taken in the UVM
        """
        for child in self.__children:
            yield self.__children[child]

    def __repr__(self):
        return self.get_full_name()

    def get_child(self, name):
        """
        13.1.3.4
        :param self:
        :param name: child string
        :return: uvm_component of that name
        """
        assert (isinstance(name, str))
        try:
            return self.__children[name]
        except KeyError:
            return None

    def get_num_children(self):
        """
        13.1.3.5
        :param self:
        :return: Number of children in component
        """
        return len(self.__children)

    def has_child(self, name):
        """
        13.1.3.6
        :param self:
        :param name: Name of child object
        :return: True if exists, False otherwise
        """
        assert (isinstance(name, str))
        return name in self.__children

    def lookup(self, name):
        """
        13.1.3.7
        Return a component base on the path. If . then
        use full name from root otherwise relative
        :param name: The search name
        :return: either the component or None
        """
        assert (isinstance(name, str))
        if name[0] == '.':
            lookup_name = name[1:]
        else:
            lookup_name = f'{self.get_full_name()}.{name}'
        try:
            return uvm_component.component_dict[lookup_name]
        except KeyError:
            return None

    def get_depth(self):
        """
        13.1.3.8
        Get the depth that I am from the top component. uvm_root is 0.
        :param self: That's me
        :return: depth
        """
        # Rather than getting all recursive just count
        # levels in the full name.
        fullname = self.get_full_name()
        if len(fullname) == 0:
            return 0
        else:
            return len(self.get_full_name().split("."))

    # noinspection SpellCheckingInspection
    def set_logging_level_hier(self, logging_level):
        """
        Set the logging level for this component's logger
        and all the way down the hierarchy
        :param logging_level: typically a constant from logging module
        :return: None
        """
        self.set_logging_level(logging_level)
        for child in self.children:
            child.set_logging_level_hier(logging_level)

    def add_logging_handler_hier(self, handler):
        """
        Add an additional handler all the way down the component hierarchy
        :param handler: A logging.Handler object
        :return: None
        """
        assert isinstance(handler, logging.Handler), \
            f"You can only add logging.Handler objects not {type(handler)}"
        self.add_logging_handler(handler)
        for child in self.children:
            child.add_logging_handler_hier(handler)

    def remove_logging_handler_hier(self, handler):
        """
        Remove a handler from all loggers below this component
        :param handler: logging handler
        :return: None
        """
        assert isinstance(handler, logging.Handler), \
            f"You must pass a logging.Handler not {type(handler)}"
        self.logger.removeHandler(handler)
        for child in self.children:
            child.remove_logging_handler_hier(handler)

    def set_formatter_on_handlers_hier(self, formatter):
        self.set_formatter_on_handlers(formatter)
        for child in self.children:
            child.set_formatter_on_handlers_hier(formatter)

    def build_phase(self):
        ...

    def connect_phase(self):
        ...

    def end_of_elaboration_phase(self):
        ...

    def start_of_simulation_phase(self):
        ...

    async def run_phase(self):
        ...

    def extract_phase(self):
        ...

    def check_phase(self):
        ...

    def report_phase(self):
        ...

    def final_phase(self):
        ...

    """
    The following sections have been skipped and could
    be implemented at a later time if necessary.
    13.1.4.2--Run time phases
    13.1.4.3--phase_* methods
    13.1.4.4--*_domain methods
    13.1.5.5--suspend
    13.1.4.6--pre_abort
    13.1.5--Configuration interface
    13.1.6--Recording interface
    13.1.7--Other interfaces
    """


class uvm_root(uvm_component, metaclass=utility_classes.UVM_ROOT_Singleton):
    """
    F.7.  We do not use uvm_pkg to hold uvm_root.  Instead it
    is a class variable of uvm_component.  This avoids
    circular reference issues regarding uvm_pkg.

    Plus, it's cleaner.

    uvm_root is still a singleton that you access through its
    constructor instead of through a get() method.

    Much of the functionality in Annex F delivers functionality
    in SystemVerilog that is already built into Python. So we're
    going to skip much of that Annex.
    """

    @classmethod
    def clear_singletons(cls):
        cls.singleton = None

    def __init__(self):
        super().__init__("uvm_root", None)
        self.uvm_test_top = None
        self.running_phase = None

    def _utt(self):
        """Used in testing"""
        return self.get_child("uvm_test_top")

    async def run_test(self, test_name=""):
        """
        This implementation skips much of the state-setting and
        what not in the LRM and focuses on building the
        hierarchy and running the test.

        At this time pyuvm has not implemented the phasing
        system described in the LRM.  It's not clear that anyone
        is using it, and in fact it is recommended that people
        stick to the basic phases.  So this implementation loops
        through the hierarchy and runs the phases.

        :param test_name: The uvm test name
        :return: none
        """

        factory = uvm_factory()
        self.uvm_test_top = factory.create_component_by_name(
            test_name, "", "uvm_test_top", self)
        try:
            for self.running_phase in uvm_common_phases:
                self.running_phase.traverse(self.uvm_test_top)
                if self.running_phase == uvm_run_phase:
                    await utility_classes.ObjectionHandler().run_phase_complete()  # noqa: E501
        except error_classes.UVMError as uve:
            self.logger.error(uve)


# In the SystemVerilog UVM the uvm_config_db is a
# convenience layer on top of the much more complicated
# uvm_resource_db class. However, few people
# use the uvm_resource_db and in fact, Mentor recommends
# that people use the uvm_config_db interface.
#
# Therefore pyuvm implements only the behavior of
# the ConfigDB. It also does NOT implement the
# regular expression form of wildcards, which are
# also rarely used. Instead it implements globbing
# as defined by fnmatch().
#
# To avoid confusion with the full uvm_config_db
# our class is named ConfigDB.
class ConfigDB(metaclass=utility_classes.Singleton):
    default_precedence = 1000
    legal_chars = set(string.ascii_letters) | set(string.digits) | set("_.")
    """
    A path-based singleton storage system
    """

    # The ConfigDB is a dual-level dict. The outer dict is a
    # GlobPathDict that can store globs in keys that match
    # later retrievals. Each entry contains another dict
    # that stores items by keys.
    #
    # Unlike the UVM config_db, this config_db makes no effort
    # to store multiple items at one location. The last stored
    # wins
    #
    # Also, pyuvm does not support wildcards in the field names
    # at this time.

    def __init__(self):
        self._path_dict = {}
        self.is_tracing = False
        self._cond_dict = {}

    def clear(self):
        """Reset the ConfigDB. Used for testing."""
        self._path_dict = {}

    @staticmethod
    def _get_context_inst_name(context, inst_name):
        """
        Get the config_key from context and passed inst_name
        :param context: uvm_component or None
        :param inst_name: string that can be a glob
        :return: string that is the key
        """
        assert context is None or isinstance(context, uvm_component), \
            "config_db context must be None or a uvm_component. "
        f"Not {type(context)}"
        if context is None:
            context = uvm_root()

        if inst_name is None or inst_name == "":
            inst_name = context.get_full_name()
        elif context.get_full_name() != "":
            inst_name = context.get_full_name() + "." + inst_name
        return context, inst_name

    def trace(self, method, context, inst_name, field_name, value):
        if self.is_tracing:
            # noinspection SpellCheckingInspection
            print(f"CFGDB/{method} Context: {context}  --  {inst_name} {field_name}={value}")  # noqa: E501

    def set(self, context, inst_name, field_name, value):
        """
        Stores an object in the db using the context and
        inst_name to create a retrieval path, and the field
        name.
        :param context: A handle to a component
        :param inst_name: The instance name within the component
        :param field_name: The field we're setting
        :param value: The object to be stored
        :return: None
        """

        if not set(field_name).issubset(self.legal_chars):
            raise error_classes.UVMNotImplemented(
                f"pyuvm does not allow wildcards in field names ({field_name})"
            )

        context, inst_name = self._get_context_inst_name(context, inst_name)

        if inst_name not in self._path_dict:
            self._path_dict[inst_name] = {}

        if field_name not in self._path_dict[inst_name]:
            self._path_dict[inst_name][field_name] = {}

        precedence = self.default_precedence
        if uvm_root().running_phase is uvm_build_phase:
            precedence = self.default_precedence - context.get_depth()

        self._path_dict[inst_name][field_name][precedence] = value

        self.trace("SET", context, inst_name, field_name, value)

    def get(self, context, inst_name, field_name):
        """
        The component path matches against the paths in the ConfigDB. The path
        cannot have wildcards, but can match against keys with wildcards.
        Return the value stored at key. Raise UVMConfigError if there is no key

        :param inst_name: component full path with no wildcards
        :param field_name: the field_name being retrieved
        :param context: The component making the call
        :return: value found at location
        """
        if not set(inst_name).issubset(self.legal_chars):
            raise error_classes.UVMError(
                f'"{inst_name}" is illegal: '
                f'inst_name wildcards only allowed when storing.')

        context, inst_name = self._get_context_inst_name(context, inst_name)

        key_matches = []  # Make the linter happy by setting this.
        try:
            # key_matches = [dk for dk in self._path_dict.keys()
            #                if fnmatch.fnmatch(inst_name, dk)]
            for dk in self._path_dict.keys():
                if fnmatch.fnmatch(inst_name, dk):
                    key_matches.append(dk)

        except TypeError:
            raise error_classes.UVMConfigItemNotFound(
                f'"{inst_name}" is not a in ConfigDB().')
        finally:
            if len(key_matches) == 0:
                raise error_classes.UVMConfigItemNotFound(
                    f'"{inst_name}" is not a in ConfigDB().')
        # Here we sort the list of paths by which paths are "in" other
        # paths. That is A comes before '*'  A.B comes before A.*, etc.
        # We use an insertion sort. A path is inserted in front of the
        # first path it is "in"
        sorted_paths = []
        try:
            sorted_paths.append(key_matches.pop())
        except IndexError:
            raise error_classes.UVMConfigItemNotFound(
                f"{inst_name} not in ConfigDB()")

        # Sort the matching keys from most specific to
        # most greedy. A.B.C before A.B.* before A.* before *
        for path in key_matches:
            inserted = False
            for ii in range(len(sorted_paths)):
                if fnmatch.fnmatch(path, sorted_paths[ii]):
                    sorted_paths.insert(ii, path)
                    inserted = True
                    break
            if not inserted:
                sorted_paths.append(path)
        value = None
        for path in sorted_paths:
            try:
                component_fields = self._path_dict[path]
                matching_path_fields = component_fields[field_name]
                max_precedence = max(matching_path_fields.keys())
                value = matching_path_fields[max_precedence]
                break
            except KeyError:
                pass
        if value is not None:
            self.trace("GET", context, inst_name, field_name, value)
            return value
        else:
            raise error_classes.UVMConfigItemNotFound(
                f'"Component {inst_name} has no field: {field_name}')

    def exists(self, context, inst_name, field_name):
        """
        Returns true if there is data in the database at this location
        :param context: None or uvm_component
        :param inst_name: instance name string in context
        :param field_name: field name for location
        :return: True if exists
        """
        try:
            _ = self.get(context, inst_name, field_name)
        except error_classes.UVMConfigItemNotFound:
            return False
        return True

    def wait_modified(self):
        raise error_classes.UVMNotImplemented(
            "wait_modified not implemented pending requests for it.")

    def __str__(self):
        str_list = [f"\n{'PATH':20}: {'FIELD':10}: {'DATA':30}"]
        for inst_path in self._path_dict:
            for field in self._path_dict[inst_path]:
                str_list.append(f"{inst_path:20}: {field:10}: {self._path_dict[inst_path][field]}")  # noqa: E501
        return "\n".join(str_list)



# The UVM TLM class hierarchy wraps the behavior of objects (usually Queues) in
# a set of classes and objects.
#
# The UVM LRM calls for three levels of classes to
# implement TLM: ports, exports, and imp:
# (from 12.2.1)
#
# * ports---Instantiated in components that require or use the associated
# interface to initiate transaction requests.
#
# * exports---Instantiated by components that forward an implementation of
# the methods defined in the associated interface.
#
# * imps---Instantiated by components that provide an implementation of or
# directly implement the methods defined.
#
# These three levels are unnecessary for Python because Python has
# multiple inheritances and duck typing.  You don't need to define
# different ports for each type being transferred.
#
# Ports have a data member named an export that implements the port
# functionality. uvm_put_port.put() calls its export.put(). Ports get
# their various flavors through multiple inheritance.


from pyuvm.s13_uvm_component import uvm_component
from pyuvm.error_classes import UVMTLMConnectionError
from pyuvm.utility_classes import UVMQueue, FIFO_DEBUG
from cocotb.queue import QueueEmpty, QueueFull
from pyuvm import error_classes
import logging


# 12.2.2
# This section describes a variety of classes without
# giving each one its own section using * to represent
# a variety of implementations (get, put, blocking_put, etc)

# Python provides multiple inheritance and we'll use that below
# to smooth implementation. Python does not require that we
# repeat the __init__ for classes that do not change the
# __init__ functionality.


# 12.2.5
# Port Classes
#
# The following port classes can be connected to "export" classes.
# They check that the export is of the correct type.
#
# uvm_port_base adds the correct methods to this class
# rather than reference them because Python allows
# you to assign functions to objects dynamically.

# We use these classes to check the connect phase
# to avoid illegal connections


class uvm_port_base(uvm_component):
    """
    A uvm_port_base is a uvm_component with a connect() function.
    The connect function creates an __export data member that
    implements the put/get,etc methods.

    We'll build functionality from uvm_port_base to create the
    other combinations of ports through multiple inheritance.

    pyuvm will make extensive use of Pythons "ask forgiveness" policy
    If you try to use the wrong method for the port you created
    then you'll get a exception, maybe a missing attribute one, though
    we could catch that one and deliver a more useful message.

    Unlike the SV implementation of UVM we return results from get and peek
    as function call returns. This is more pythonic.
    """

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.connected_to = {}
        self.provided_to = {}
        self.export = None

    def connect(self, export):
        """
        Attach this port to the associated export.

        :param export:
        :return:
        """
        try:
            self.export = export
            self.connected_to[export.get_full_name()] = export
            export.provided_to[self.get_full_name()] = self
        except KeyError:
            raise UVMTLMConnectionError(
                f"Error connecting {self.get_name()} using {export}")

    @staticmethod
    def check_export(export, check_class):
        if not isinstance(export, check_class):
            raise UVMTLMConnectionError(
                f"{export} must be an instance "
                "of\n{check_class} not\n{type(export)}")


# put

# 12.2.5.1
class uvm_blocking_put_port(uvm_port_base):
    """
    Access the blocking put interfaces
    """

    def connect(self, export):
        self.check_export(export, uvm_blocking_put_port)
        super().connect(export)

    # 12.2.4.2.1
    async def put(self, datum):
        """
         A blocking put that calls the export.put
        :param datum: Datum to put
        :return: None
        """
        try:
            await self.export.put(datum)
        except AttributeError:
            raise UVMTLMConnectionError(
                "Missing or wrong export in"
                f" {self.get_full_name()}. Did you connect it?")


# 12.2.5.1
class uvm_nonblocking_put_port(uvm_port_base):
    """
    Access the non_blocking put interface
    """

    def connect(self, export):
        self.check_export(export, uvm_nonblocking_put_port)
        super().connect(export)

    # 12.2.4.2.4
    def try_put(self, data):
        """
        Tries to put data on a port, but if the
        port is full it returns False

        :param data: data to deliver
        :return: True = success
        """
        try:
            success = self.export.try_put(data)
            return success
        except AttributeError:
            raise UVMTLMConnectionError(
                "Missing or wrong export "
                f"in {self.get_full_name()}. Did you connect it?")

    # 12.2.4.2.5
    def can_put(self):
        """
        Returns true if there is room for data to
        be put on the port

        :return: bool
        """
        try:
            can_do_it = self.export.can_put()
            return can_do_it
        except AttributeError:
            raise UVMTLMConnectionError(
                "Missing or wrong export in"
                f" {self.get_full_name()}. Did you connect it?")


# 12.2.5.1
# The put port delivers blocking puts and gets.
# Here is the multiple inheritance that SV lacked
class uvm_put_port(uvm_blocking_put_port, uvm_nonblocking_put_port):
    ...


# 12.2.5.1
class uvm_blocking_get_port(uvm_port_base):
    """
        Access the blocking get export methods
    """

    def connect(self, export):
        self.check_export(export, uvm_blocking_get_port)
        super().connect(export)

    # 12.2.4.2.2
    async def get(self):
        """

        A blocking get that returns the data got
        :return: data
        """
        try:
            data = await self.export.get()
            return data
        except AttributeError:
            raise UVMTLMConnectionError(
                "Missing or wrong export in "
                f"{self.get_full_name()}. Did you connect it?")


# 12.2.5.1
class uvm_nonblocking_get_port(uvm_port_base):
    """
    Access the non_blocking methods in export
    """

    def connect(self, export):
        self.check_export(export, uvm_nonblocking_get_port)
        super().connect(export)

    def try_get(self):
        """
        12.2.4.2.6
        Returns a tuple containing success and the data
        This is different than SV UVM that returns the
        data through the argument list.
        :return: (success, data)
        """

        try:
            (success, data) = self.export.try_get()
        except AttributeError:
            raise UVMTLMConnectionError(
                "Missing or wrong export in"
                f" {self.get_full_name()}. Did you connect it?")

        return success, data

    # 12.2.4.2.7
    def can_get(self):
        """
        Returns true if there is data to get

        :return: bool
        """
        try:
            can = self.export.can_get()
        except AttributeError:
            raise UVMTLMConnectionError(
                "Missing or wrong export in"
                f" {self.get_full_name()}. Did you connect it?")

        return can


class uvm_get_port(uvm_blocking_get_port, uvm_nonblocking_get_port):
    ...


#
# 12.2.5.1
class uvm_blocking_peek_port(uvm_port_base):
    """
    Provides access to the peek methods
    """

    def connect(self, export):
        self.check_export(export, uvm_blocking_peek_port)
        super().connect(export)

    # 12.2.4.2.3
    async def peek(self):
        """
        A blocking peek that returns data without
        consuming it.

        :return: datum
        """
        try:
            datum = await self.export.peek()
            return datum
        except AttributeError:
            raise UVMTLMConnectionError(
                "Missing or wrong export in"
                f" {self.get_full_name()}. Did you connect it?")


# 12.2.5.1
class uvm_nonblocking_peek_port(uvm_port_base):
    """
    Try a peek
    """

    def connect(self, export):
        self.check_export(export, uvm_nonblocking_peek_port)
        super().connect(export)

    # 12.2.4.2.8
    def try_peek(self):
        """

        Tries to peek for data and returns
        a tuple with success and the data
        :return: (success, data)
        """
        try:
            success, data = self.export.try_peek()
        except AttributeError:
            raise UVMTLMConnectionError(
                "Missing or wrong export in"
                f" {self.get_full_name()}. Did you connect it?")
        return success, data

    # 12.2.4.2.9
    def can_peek(self):
        """
        Checks if peeking will be successful

        :return: True if can peek
        """
        can = self.export.can_peek()
        return can


class uvm_peek_port(uvm_blocking_peek_port, uvm_nonblocking_peek_port):
    ...


# get_peek

class uvm_blocking_get_peek_port(uvm_blocking_get_port,
                                 uvm_blocking_peek_port):
    def connect(self, export):
        if not (isinstance(export, uvm_blocking_get_port) or isinstance(export, uvm_blocking_peek_port)):  # noqa: E501
            raise UVMTLMConnectionError(
                f"Tried to connect {type(export)} to "
                f"uvm_blocking_get_peek {self.get_full_name()}")
        super().connect(export)


class uvm_nonblocking_get_peek_port(uvm_nonblocking_get_port,
                                    uvm_nonblocking_peek_port):
    def connect(self, export):
        if not (isinstance(export, uvm_nonblocking_get_port) or isinstance(export, uvm_nonblocking_peek_port)):  # noqa: E501
            raise UVMTLMConnectionError(
                f"Tried to connect an {export} to "
                f"uvm_blocking_get_peek {self.get_full_name()}")
        super().connect(export)


class uvm_get_peek_port(uvm_blocking_get_peek_port,
                        uvm_nonblocking_get_peek_port):
    def connect(self, export):
        if not (isinstance(export, uvm_nonblocking_get_port) or isinstance(export, uvm_nonblocking_peek_port)  # noqa: E501
                or isinstance(export, uvm_blocking_get_port) or isinstance(export, uvm_blocking_peek_port)):  # noqa: E501, W503
            raise UVMTLMConnectionError(
                "Tried to connect an illegal "
                f"export to uvm_blocking_get_peek {self.get_full_name()}")
        super().connect(export)


class uvm_blocking_transport_port(uvm_port_base):
    def __init__(self, name, parent):
        super().__init__(name, parent)

    def connect(self, export):
        self.check_export(export, uvm_blocking_transport_port)
        super().connect(export)

    async def transport(self, put_data):
        try:
            get_data = await self.export.transport(put_data)
        except AttributeError:
            raise UVMTLMConnectionError(
                "Missing or wrong export in"
                f" {self.get_full_name()}. Did you connect it?")
        return get_data


class uvm_nonblocking_transport_port(uvm_port_base):

    def __init__(self, name, parent):
        super().__init__(name, parent)

    def connect(self, export):
        self.check_export(export, uvm_nonblocking_transport_port)
        super().connect(export)

    def nb_transport(self, put_data):
        try:
            success, get_data = self.export.nb_transport(put_data)
        except AttributeError:
            raise UVMTLMConnectionError(
                "Missing or wrong export in"
                f" {self.get_full_name()}. Did you connect it?")
        return success, get_data


class uvm_transport_port(uvm_blocking_transport_port,
                         uvm_nonblocking_transport_port):
    ...


# master
class uvm_blocking_master_port(uvm_blocking_put_port,
                               uvm_blocking_get_peek_port):
    ...


class uvm_nonblocking_master_port(uvm_nonblocking_put_port,
                                  uvm_nonblocking_get_peek_port):
    ...


class uvm_master_port(uvm_blocking_master_port, uvm_nonblocking_master_port):
    ...


class uvm_blocking_slave_port(uvm_blocking_put_port,
                              uvm_blocking_get_peek_port):
    ...


class uvm_nonblocking_slave_port(uvm_nonblocking_get_peek_port,
                                 uvm_nonblocking_put_port):
    ...


class uvm_slave_port(uvm_nonblocking_slave_port,
                     uvm_blocking_slave_port):
    ...


class uvm_analysis_port(uvm_port_base):
    def __init__(self, name, parent):
        super().__init__(name, parent)

        self.subscribers = []

    def connect(self, export):
        self.check_export(export, uvm_analysis_port)
        self.subscribers.append(export)

    # 12.2.8.1
    def write(self, datum):
        """
        :param datum: data to send
        :return: None
        """
        for export in self.subscribers:
            if not hasattr(export, "write"):
                raise UVMTLMConnectionError(
                    f"No write() method in {export}. Did you connect it?")
            export.write(datum)


# THE UNNECESSARY EXPORTS
# It turns out the entire export class is unnecessary in Python
# Since the port's connect() method can connect to either a port
# or export (they both provided the needed methods) it means
# that an export is simply a port with the methods filled in.
#
# One could have made ports abstract base classes, but that
# would add complications to the metatypes that implement
# the factory, so we don't do that.

# We provide the export class below to follow the spec
# and to acknowledge that the name "export" does imply
# a port with implemented functions.

class uvm_nonblocking_put_export(uvm_nonblocking_put_port):
    ...


class uvm_blocking_put_export(uvm_blocking_put_port):
    ...


class uvm_put_export(uvm_nonblocking_put_export, uvm_blocking_put_export):
    ...


# get
class uvm_nonblocking_get_export(uvm_nonblocking_get_port):
    ...


class uvm_blocking_get_export(uvm_blocking_get_port):
    ...


class uvm_get_export(uvm_blocking_get_export, uvm_nonblocking_get_export):
    ...


# peek
class uvm_nonblocking_peek_export(uvm_nonblocking_peek_port):
    ...


class uvm_blocking_peek_export(uvm_blocking_peek_port):
    ...


class uvm_peek_export(uvm_nonblocking_peek_export, uvm_blocking_peek_export):
    ...


# get_peek
class uvm_blocking_get_peek_export(uvm_blocking_get_peek_port):
    ...


class uvm_nonblocking_get_peek_export(uvm_nonblocking_get_peek_port):
    ...


class uvm_get_peek_export(uvm_nonblocking_get_peek_export,
                          uvm_blocking_get_peek_export):
    ...


# transport
class uvm_blocking_transport_export(uvm_blocking_transport_port):
    ...


class uvm_nonblocking_transport_export(uvm_nonblocking_transport_port):
    ...


class uvm_transport_export(uvm_nonblocking_transport_export,
                           uvm_blocking_transport_export):
    ...


# master
class uvm_blocking_master_export(uvm_blocking_put_export,
                                 uvm_blocking_get_peek_export):
    ...


class uvm_nonblocking_master_export(uvm_blocking_peek_export,
                                    uvm_nonblocking_get_peek_export):
    ...


class uvm_master_export(uvm_blocking_master_export,
                        uvm_nonblocking_master_export):
    ...


# slave
class uvm_blocking_slave_export(uvm_blocking_put_export,
                                uvm_blocking_get_peek_export):
    ...


class uvm_nonblocking_slave_export(uvm_nonblocking_put_export,
                                   uvm_nonblocking_get_peek_export):
    ...


class uvm_slave_export(uvm_blocking_slave_export,
                       uvm_nonblocking_slave_export):
    ...


class uvm_analysis_export(uvm_analysis_port):
    """
    The analysis export overrides the port's write method and forces others
    to override its write method.
    """

    def write(self, data):
        raise error_classes.UVMTLMConnectionError(
            "If you extend uvm_analysis_export, or uvm_subscriber, you must"
            " override the write method")


# 12.2.8 FIFO Classes
#
# These classes provide synchronization control between
# threads using the Queue class.
#
# One note.  The RLM has only 12.2.8.1.3 and 12.2.8.1.4, put_export
# and get_peek_export, but the UVM code has all the variants
# of exports.
#
# The SystemVerilog UVM relies upon static type checking
# and polymorphism to make sure that users connect the
# correct ports to these exports, but we don't have
# static checking, so we implement classes for all the
# port variants. This creates the runtime assertion checking
# that we need.


class QueueAccessor:
    def __init__(self, name, parent, uvm_queue, ap):
        super(QueueAccessor, self).__init__(name, parent)
        assert (isinstance(uvm_queue, UVMQueue)), \
            "Tried to pass a non-UVMQueue to QueueAccessor constructor"
        self.queue = uvm_queue
        self.ap = ap


class uvm_tlm_fifo_base(uvm_component):
    """
    Declares and instantiate the exports needed to communicate
    through the Queue.
    """

    class BlockingPutExport(QueueAccessor, uvm_blocking_put_export):
        async def put(self, item):
            self.logger.log(FIFO_DEBUG, f"blocking put: {item}")
            await self.queue.put(item)
            self.logger.log(FIFO_DEBUG, f"success put {item}")
            self.ap.write(item)

    #  12.2.8.1.3
    class NonBlockingPutExport(QueueAccessor, uvm_nonblocking_put_export):

        def can_put(self):
            return not self.queue.full()

        def try_put(self, item):
            try:
                self.queue.put_nowait(item)
                self.ap.write(item)
                return True
            except QueueFull:
                return False

    class PutExport(BlockingPutExport, NonBlockingPutExport):
        ...

    class BlockingGetExport(QueueAccessor, uvm_blocking_get_export):
        async def get(self):
            self.logger.log(FIFO_DEBUG, "Attempting blocking get")
            item = await self.queue.get()
            self.logger.log(FIFO_DEBUG, f"got {item}")
            self.ap.write(item)
            return item

    class NonBlockingGetExport(QueueAccessor, uvm_nonblocking_get_export):
        def can_get(self):
            get_ok = not self.queue.empty()
            return get_ok

        def try_get(self):
            try:
                item = self.queue.get_nowait()
                self.ap.write(item)
                return True, item
            except QueueEmpty:
                return False, None

    class GetExport(BlockingGetExport, NonBlockingGetExport):
        ...

    class BlockingPeekExport(QueueAccessor, uvm_blocking_peek_export):
        async def peek(self):
            self.logger.log(FIFO_DEBUG, "Attempting blocking peek")
            peek_data = await self.queue.peek()
            self.logger.log(FIFO_DEBUG, f"peeked at {peek_data}")
            return peek_data

    class NonBlockingPeekExport(QueueAccessor, uvm_nonblocking_peek_export):
        def can_peek(self):
            return not self.queue.empty()

        def try_peek(self):
            try:
                datum = self.queue.peek_nowait()
                return True, datum
            except QueueEmpty:
                return False, None

    class PeekExport(BlockingPeekExport, NonBlockingPeekExport):
        ...

    class BlockingGetPeekExport(BlockingGetExport, BlockingPeekExport):
        ...

    class NonBlockingGetPeekExport(NonBlockingGetExport,
                                   NonBlockingPeekExport):
        ...

    #  12.2.8.1.4
    class GetPeekExport(GetExport, PeekExport):
        ...

    def __init__(self, name, parent, maxsize=1):
        super().__init__(name, parent)
        self.queue = UVMQueue(maxsize=maxsize)
        self.get_ap = uvm_analysis_port("get_ap", self)
        self.put_ap = uvm_analysis_port("put_ap", self)

        self.blocking_put_export = self.BlockingPutExport("blocking_put_export",  # noqa: E501
                                                          self, self.queue,
                                                          self.put_ap)
        self.nonblocking_put_export = self.NonBlockingPutExport("nonblocking_put_export",  # noqa: E501
                                                                self, self.queue, self.put_ap)  # noqa: E501
        self.put_export = self.PutExport("put_export", self, self.queue, self.put_ap)  # noqa: E501

        self.blocking_get_export = self.BlockingGetExport("blocking_get_export", self,  # noqa: E501
                                                          self.queue, self.get_ap)  # noqa: E501
        self.nonblocking_get_export = self.NonBlockingGetExport("nonblocking_get_export", self,  # noqa: E501
                                                                self.queue, self.put_ap)  # noqa: E501

        self.get_export = self.GetExport("get_export", self, self.queue,
                                         self.get_ap)

        self.blocking_peek_export = self.BlockingPeekExport("blocking_peek_export", self,  # noqa: E501
                                                            self.queue, self.get_ap)  # noqa: E501
        self.nonblocking_peek_export = self.NonBlockingPeekExport("nonblocking_peek_export", self,  # noqa: E501
                                                                  self.queue, self.get_ap)  # noqa: E501
        self.peek_export = self.PeekExport("peek_export", self,
                                           self.queue, self.get_ap)

        self.blocking_get_peek_export = self.BlockingGetPeekExport("blocking_get_peek_export", self,  # noqa: E501
                                                                   self.queue, self.get_ap)  # noqa: E501
        self.nonblocking_get_peek_export = self.NonBlockingGetPeekExport("nonblocking_get_peek_export", self,  # noqa: E501
                                                                         self.queue, self.get_ap)  # noqa: E501
        self.get_peek_export = self.GetPeekExport("get_peek_export", self, self.queue, self.get_ap)  # noqa: E501

    def end_of_elaboration_phase(self):
        formatter = logging.Formatter("%(levelname)s: [%(name)s]: %(message)s")
        self.set_formatter_on_handlers_hier(formatter)


class uvm_tlm_fifo(uvm_tlm_fifo_base):

    #  12.2.8.2.1
    def __init__(self, name=None, parent=None, size=1):
        """ uvm_tlm_fifo is a uvm_component"""
        super().__init__(name, parent, size)

    # 12.2.8.2.2
    def size(self):
        """
        Return the size of the fifo

        :return: size of FIFO
        """
        return self.queue.maxsize

    # 12.2.8.2.3
    def used(self):
        """
        How much of the FIFO is being used?

        :return: Size of the FIFO
        """
        return self.queue.qsize()

    # 12.2.8.2.4
    def is_empty(self):
        """
        Returns true if FIFO is empty

        :return: True if empty
        """
        return self.queue.empty()

    # 12.2.8.2.5
    def is_full(self):
        """
        Test for full FIFO

        :return: True if full
        """
        return self.queue.full()

    #  12.2.8.2.6
    #     The SystemVerilog UVM flushes the Queue
    #     using a while loop and counting the number
    #     of threads waiting for a get. If there are
    #     still threads waiting after the loop runs
    #     it issues an error.
    #
    #     In pyuvm we just flush the loop and clear
    #     all the unfinished tasks.

    def flush(self):
        """
        Flush out the FIFO
        """
        self.queue._queue.clear()


class uvm_tlm_analysis_fifo(uvm_tlm_fifo):
    class AnalysisExport(QueueAccessor, uvm_analysis_port):
        def write(self, item):
            try:
                self.queue.put_nowait(item)
            except QueueFull:
                raise QueueFull(
                    f"Full analysis fifo: {self.__name__}. This should never happen")  # noqa: E501

    def __init__(self, name, parent=None):
        super().__init__(name, parent, 0)
        self.analysis_export = self.AnalysisExport(name="analysis_export",
                                                   parent=self,
                                                   uvm_queue=self.queue,
                                                   ap=None)


#    12.2.9.1
class uvm_tlm_req_rsp_channel(uvm_component):
    class MasterSlaveExport(uvm_master_port, uvm_get_peek_port):
        def __init__(self, name, parent, put_export, get_peek_export):
            super().__init__(name, parent)
            self.put_export = put_export
            self.get_peek_export = get_peek_export

        async def put(self, item):
            await self.put_export.put(item)

        def can_put(self):
            return self.put_export.can_put()

        def try_put(self, item):
            return self.put_export.try_put(item)

        async def get(self):
            return await self.get_peek_export.get()

        def can_get(self):
            return self.get_peek_export.can_get()

        def try_get(self):
            return self.get_peek_export.try_get

    def __init__(self, name, parent=None, request_fifo_size=1,
                 response_fifo_size=1):
        super().__init__(name, parent)
        self.req_tlm_fifo = uvm_tlm_fifo("request_fifo", self,
                                         request_fifo_size)
        self.rsp_tlm_fifo = uvm_tlm_fifo("response_fifo", self,
                                         response_fifo_size)

        self.put_request_export = self.req_tlm_fifo.put_export
        self.get_peek_response_export = self.rsp_tlm_fifo.get_peek_export
        self.get_peek_request_export = self.req_tlm_fifo.get_peek_export
        self.put_response_export = self.rsp_tlm_fifo.put_export
        self.request_ap = uvm_analysis_port("request_ap", self)
        self.response_ap = uvm_analysis_port("response_ap", self)

        self.master_export = self.MasterSlaveExport(
            name="master_export",
            parent=self,
            put_export=self.put_request_export,
            get_peek_export=self.get_peek_response_export)
        self.slave_export = self.MasterSlaveExport(
            name="slave_export",
            parent=self,
            get_peek_export=self.get_peek_request_export,
            put_export=self.put_response_export)

    def connect_phase(self):
        self.request_ap.connect(self.req_tlm_fifo.put_ap)
        self.response_ap.connect(self.rsp_tlm_fifo.get_ap)


class uvm_tlm_transport_channel(uvm_tlm_req_rsp_channel):
    class TransportExport(uvm_transport_port):
        def __init__(self, name, parent, req_fifo, rsp_fifo):
            super().__init__(name, parent)
            self.req_fifo = req_fifo
            self.rsp_fifo = rsp_fifo

        async def transport(self, req):
            self.req_fifo.put_export.put(req)
            return await self.rsp_fifo.get_peek_export.get()

        def nb_transport(self, req):
            if not self.req_fifo.put_export.try_put(req):
                return False
            return self.rsp_fifo.get_peek_export.try_get()

    def __init__(self, name, parent=None):
        super().__init__(name, parent, 1, 1)
        self.transport_export = self.TransportExport(
            "transport_export", self,
            req_fifo=self.req_tlm_fifo,
            rsp_fifo=self.rsp_tlm_fifo)

# UVM TLM 2
# 12.3

# This is left for future development.



from enum import Enum, auto
# Support Modules
from pyuvm.error_classes import *
from pyuvm.utility_classes import *
# Section 5
from pyuvm.s05_base_classes import *
# Section 6
from pyuvm.s06_reporting_classes import *
# Section 8
from pyuvm.s08_factory_classes import *
# Section 9
from pyuvm.s09_phasing import *
# Section 12
from pyuvm.s12_uvm_tlm_interfaces import *
# Section 13
from pyuvm.s13_predefined_component_classes import *
# Section 14, 15 (Done as fresh Python design)
from pyuvm.s14_15_python_sequences import *







# from base_classes import *
from pyuvm.s14_15_python_sequences import uvm_seq_item_port
from pyuvm.s12_uvm_tlm_interfaces import uvm_analysis_export
from pyuvm.s13_uvm_component import *
from pyuvm import error_classes
from enum import IntEnum


# This section and sequences are the crux of pyuvm.
# The classes here allow us to build classic UVM
# testbenches in Python.
#
# Section 13 of the IEEE-UVM Reference Manual (1800.2-2017)
# lists five pieces of uvm_component functionality.
# pyuvm implements much of this functionality using Python:
#
# a: Hierarchy---This is implemented in pyuvm
# b: Phasing---This is also implemented in pyuvm, but is hardcoded
#              to the standard UVM phases.
# c: Hierarchical Reporting---We manage this with the logging module. It is
#              orthogonal to the components.
# d: Transaction Recording---We do not record transactions since pyuvm does
#              not run in the simulator. This
# could be added later if we see a need or way to do it.
# e: Factory---pyuvm manages the factory through the create() method without
#              all the SystemVerilog typing overhead.
#


class uvm_active_passive_enum(IntEnum):
    UVM_PASSIVE = 0
    UVM_ACTIVE = 1


# Class Declarations


# 13.2
class uvm_test(uvm_component):
    """
    The base class for all user-defined tests

    Python does not require that we override new() for every
    UVM class, so we don't do that (same for __init__).
    """


# 13.3
class uvm_env(uvm_component):
    """
    The user's contains for agents and what-not
    """


# 13.4
class uvm_agent(uvm_component):
    """
    Contains controls for individual agents
    """

    def build_phase(self):
        super().build_phase()
        self.is_active = uvm_active_passive_enum.UVM_ACTIVE
        try:
            self.is_active = self.cdb_get("is_active")
        except error_classes.UVMConfigItemNotFound:
            self.is_active = uvm_active_passive_enum.UVM_ACTIVE

        if self.is_active not in list(uvm_active_passive_enum):
            self.logger.warning(f"{self.get_full_name()}"
                                "has illegal is_active"
                                f" value: {self.is_active}."
                                "Setting to UVM_ACTIVE")
            self.is_active = uvm_active_passive_enum.UVM_ACTIVE

    def get_is_active(self):
        return self.is_active

    def active(self):
        return self.get_is_active() == uvm_active_passive_enum.UVM_ACTIVE


# 13.5
class uvm_monitor(uvm_component):
    ...


# 13.6
class uvm_scoreboard(uvm_component):
    ...


# 13.7
class uvm_driver(uvm_component):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.seq_item_port = uvm_seq_item_port("seq_item_port", self)
        pass


# 13.8 uvm_push_driver

# Never seen one used. Not implemented.


# 13.9
class uvm_subscriber(uvm_analysis_export):
    ...



from collections import OrderedDict
import logging
import fnmatch
import cocotb.queue
from cocotb.triggers import Event
from cocotb.queue import QueueEmpty

FIFO_DEBUG = 5
logging.addLevelName(FIFO_DEBUG, "FIFO_DEBUG")


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)  # noqa: E501
        return cls._instances[cls]

    @classmethod
    def clear_singletons(mcs):
        mcs._instances.clear()
        pass


class Override:
    """
    This class stores an override and an optional path.
    It is intended to be stored in a dict with the original class
    as the key.
    """

    def __init__(self):

        self.type_override = None
        self.inst_overrides = OrderedDict()

    def add(self, override, path=None):
        if path is None:
            self.type_override = override
        else:
            self.inst_overrides[path] = override

    def find_inst_override(self, path):
        for inst in self.inst_overrides:
            if fnmatch.fnmatch(path, inst):
                return self.inst_overrides[inst]
        return None

    def __str__(self):
        """
        For printing out the overrides
        :return: str
        """
        if self.type_override is not None:
            to = "Type Override: " + f"{self.type_override.__name__}"
        else:
            to = "Type Override: None"
        ss = f"{to:25}" + " || "
        if self.inst_overrides is not None:
            ss += "Instance Overrides: "
            first = True
            for inst in self.inst_overrides:
                if not first:
                    ss += " | "
                first = False
                if len(inst) > 29:
                    inst = inst[:29]
                ss += inst + f" => {self.inst_overrides[inst].__name__}"

        return ss


class FactoryData(metaclass=Singleton):

    def __init__(self):
        self.classes = {}
        self.clear_overrides()
        self.logger = logging.getLogger("Factory")

    def clear_overrides(self):
        self.overrides = {}

    def clear_classes(self):
        self.classes = {}

    # From 8.3.1.5
    def find_override(self, requested_type,
                      inst_path=None, overridden_list=None):
        """
        :param requested_type: The type we're overriding
        :param inst_path: The inst_path we're using to override if any
        :param overridden_list: A list of previously found overrides
        :return: overriding_type

        Override searches are recursively applied, with instance
        overrides taking precedence over type overrides.
        If foo overrides bar, and xyz overrides foo, then a request for bar
        returns xyx.
        """

        # xyz -> foo -> bar
        #
        # So if find_override is fo:
        #     fo(xyz) -> fo(foo) -> fo(bar) <-- no override returns bar.
        # Recursive loops result ina n error in which case the
        # type returned is the one that formed the loop:
        # xyz -> foo -> bar -> xyz
        #
        # fo(xyz) -> fo(foo) -> fo(bar) -> fo(xyz) -- xyz is in
        # list of overrides so return bar
        # bar is returned with a printed error.
        #
        # We use the Override class which contains both the type override if
        # there is one or
        # a list of instance overrides in the order the were added.
        # If inst_path is None we return the type_override or its override
        # If inst_path is given, but we don't find a match we return
        # type_override if it exists

        # Keep track of what classes have been overridden
        #

        # Is there an override loop?
        # noinspection PyShadowingNames
        def check_override(override, overridden_list):
            if overridden_list is None:
                overridden_list = []
            if override in overridden_list:
                self.logger.error(
                    f"{requested_type} already overridden: {overridden_list}")
                return requested_type
            else:
                overridden_list.append(requested_type)
                rec_override = self.find_override(override,
                                                  inst_path,
                                                  overridden_list)
                return rec_override

        # Save the type for a later check
        # Is this requested type even in the list of overrides?
        try:
            override = self.overrides[requested_type]
        except KeyError:
            return requested_type

        if inst_path is not None:
            for path in override.inst_overrides:
                if fnmatch.fnmatch(inst_path, path):
                    found_type = override.inst_overrides[path]
                    return check_override(found_type, overridden_list)

        # No inst requested or found, do we have a type override?
        if override.type_override is not None:
            return check_override(override.type_override, overridden_list)
        else:
            return requested_type


class FactoryMeta(type):
    """
    This is the metaclass that causes all uvm_void classes
    to register themselves
    """

    def __init__(cls, name, bases, cls_dict):
        FactoryData().classes[cls.__name__] = cls
        super().__init__(name, bases, cls_dict)


class uvm_void(metaclass=FactoryMeta):
    """
    5.2
    SystemVerilog Python uses this class to allow all
    uvm objects to be stored in a uvm_void variable through
    polymorphism.

    In pyuvm, we're using uvm_void() as a metaclass so
    that all UVM classes can be stored in a factory.
"""


class UVM_ROOT_Singleton(FactoryMeta):
    singleton = None

    def __call__(cls, *args, **kwargs):
        if cls.singleton is None:
            cls.singleton = super(UVM_ROOT_Singleton, cls).__call__(*args, **kwargs)  # noqa : E501
        return cls.singleton

    @classmethod
    def clear_singletons(cls):
        cls.singleton = None
        pass


class ObjectionHandler(metaclass=Singleton):
    """
    This singleton accepts objections and then allows
    them to be removed. It returns True to run_phase_complete()
    when there are no objections left.
    """

    def __init__(self):
        self.__objections = {}
        self._objection_event = Event("objection changed")
        self.objection_raised = False
        self.run_phase_done_flag = None  # used in test suites
        self.printed_warning = False

    def __str__(self):
        ss = f"run_phase complete: {self.run_phase_complete()}\n"
        ss += "Current Objections:\n"
        for cc in self.__objections:
            ss += f"{self.__objections[cc]}\n"
        return ss

    def raise_objection(self, raiser):
        self.__objections[raiser] = raiser.get_full_name()
        self.objection_raised = True
        self._objection_event.clear()

    def drop_objection(self, dropper):
        try:
            del self.__objections[dropper]
        except KeyError:
            self.objection_raised = True
            pass
        if len(self.__objections) == 0:
            self._objection_event.set()

    async def run_phase_complete(self):
        if self.objection_raised:
            await self._objection_event.wait()
        else:
            logging.warning(
                "You did not call self.raise_objection() in any run_phase")


class UVMQueue(cocotb.queue.Queue):
    """
    The UVMQueue provides a peek function as well as the
    ability to break out of a blocking operation if
    the time_to_die predicate is true.  The time
    to die is set to the dropping of all run_phase objections
    by default.
    """
    def __str__(self):
        return str(self._queue)

    def _peek(self):
        return self._queue[0]

    async def peek(self):
        """Remove and return an item from the queue.
        If the queue is empty, wait until an item is available.
        """
        while self.empty():
            event = Event('{} peek'.format(type(self).__name__))
            self._getters.append((event, cocotb.scheduler._current_task))
            await event.wait()
        return self.peek_nowait()

    def peek_nowait(self):
        """Remove and return an item from the queue.
        Return an item if one is immediately available, else raise
        :exc:`asyncio.QueueEmpty`.
        """
        if self.empty():
            raise QueueEmpty()
        item = self._peek()
        return item



# The SystemVerilog sequences provided much more functionality
# than most users ever used. This led to extremely complicated
# code.
#
# This file implements the uvm sequence functionality in
# Python using Python features instead of emulating
# SystemVerilog features.


from pyuvm.s05_base_classes import *
from pyuvm.s12_uvm_tlm_interfaces import *
from cocotb.triggers import Event as CocotbEvent

# The sequence system allows users to create and populate sequence
# items and then send them to a driver. The driver
# loops through getting the next sequence item,
# processing it, and sending the result back.
#
# Remembering that all run_phases run in their own
# thread we see this code in the driver.
#
# def run_phase(phase):
#     while True:
#        req = self.seq_item_port.get_next_item()
#        # send the req to the tinyALU and get rsp
#        self.seq_item_port.item_done(rsp)
#
# or
#    while True:
#        req = self.seq_item_port.get_next_item()
#        # do stuff
#        self.seq_item_port.item_done()
#        self.seq_item_port.put(rsp)
#
# Either way the sequence in this case does:
#
# start_item(req)
# finish_item(req)
# rsp = get_response()
#
# We see above that the driver is a simple uvm_component
# with a special port. The port does all the synchronization.
# It blocks until there is a req and then it sends the response
# back and notifies the sequencer that the transaction is done.  we have:
#
#
# From the other side, the sequence side we get this:
#
# First someone starts the sequence:
#
# test_seq.start(seqr)
#
# This puts a handle to the sequencer (seqr) in the sequence.
# Then this happens.
#
# def body():
#    req = Req()
#    self.start_item(req) # Request the sequencer
#    req.A = 1
#    req.B = 5
#    req.op = operators.ADD
#    self.finish_item(req) # Send and wait for item_done
#    rsp = self.get_response()
#
#    The above puts this sequence in a queue and blocks until
#      the sequence's turn comes up.

# So the sequence contains:
# start()
# start_item()
# finish_item()
# get_response()
#
# The seq_item_port (and export) contain:
# get_next_item()
# item_done()
# put()
#
# The sequencer that connects them contains:
# A fifo that holds sequences in order
# A mechanism to notify start_item that it's turn has arrived
# A mechanism to notify finish_item that the item is done
# A mechanism to return responses
# a seq_item_export
#
# The driver contains
# a seq_item_port
#
# We'll build from the seq_item_port out.


# uvm_seq_item_port
# The uvm_seq_item_port is a uvm_put_port with two extra methods.

class ResponseQueue(UVMQueue):
    """
    Returns either the next response or the item with the id.
    """

    def __init__(self, maxsize: int = 0):
        super().__init__(maxsize=maxsize)
        self.put_event = CocotbEvent("put event")

    def put_nowait(self, item):
        super().put_nowait(item)
        self.put_event.set()
        self.put_event.clear()

    async def get_response(self, txn_id=None):
        if txn_id is None:
            return await self.get()
        else:
            while True:
                item_list = list(self._queue)
                txn_list = [xx
                            for xx in item_list
                            if xx.transaction_id == txn_id]
                if len(txn_list) == 0:
                    await self.put_event.wait()
                else:
                    assert len(txn_list) == 1, \
                        f"Multiple transactionsn have the same ID: {txn_id}"
                    _ = self._queue.index(txn_list[0])
                    self._queue.remove(txn_list[0])
                    return txn_list[0]

    def __str__(self):
        return str([str(xx) for xx in self.queue])


class uvm_sequence_item(uvm_transaction):
    """
    The pyuvm uvm_sequence_item has conditions to
    implement start_item() and finish_item()
    """

    def __init__(self, name):
        super().__init__(name)
        self.start_condition = CocotbEvent()
        self.finish_condition = CocotbEvent()
        self.parent_sequence_id = None
        self.response_id = None

    def set_context(self, item):
        """
        Use this to link a new response transaction to the request transaction.
        rsp.set_context(req)

        :param item: The request transaction
        :return:
        """
        self.response_id = (item.parent_sequence_id, item.get_transaction_id())


class uvm_seq_item_export(uvm_blocking_put_export):
    """
    The sequence item port with a request queue and
    a response queue.
    """

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.req_q = UVMQueue()
        self.rsp_q = ResponseQueue()
        self.current_item = None

    async def put_req(self, item):
        """
        put request into request queue

        :param item: request item
        :return:
        """
        await self.req_q.put(item)

    def put_response(self, item):
        """
        Put response into response queue

        :param item: response item
        :return:
        """
        self.rsp_q.put_nowait(item)

    async def get_next_item(self):
        """
        Get the next item out of the item queue
        :return: item to process
        """
        if self.current_item is not None:
            raise error_classes.UVMSequenceError(
                "You must call item_done() before calling get_next_item again")
        self.current_item = await self.req_q.get()
        return self.current_item

    def item_done(self, rsp=None):
        """
        Signal that the item has been completed. If item is not none
        put it into the response queue.

        :param rsp: item to put in response queue if not None
        """
        if self.current_item is None:
            raise error_classes.UVMSequenceError(
                "You must call get_next_item before calling item_done")

        self.current_item.finish_condition.set()
        self.current_item.finish_condition.clear()
        self.current_item = None
        if rsp is not None:
            self.put_response(rsp)

    async def get_response(self, transaction_id=None):
        """
        If transaction_id is not none, block until a
        response with the transaction id becomes available.
        :param transaction_id: The transaction ID of the response
        :return:
        """
        datum = await self.rsp_q.get_response(transaction_id)
        return datum


class uvm_seq_item_port(uvm_blocking_put_port):
    def connect(self, export):
        self.check_export(export, uvm_seq_item_export)
        super().connect(export)

    async def put_req(self, item):
        """Put a request item in the request queue"""
        await self.export.put_req(item)

    def put_response(self, item):
        """Put a response back in the queue. aka put_response"""
        self.export.put_response(item)

    async def get_next_item(self):
        """get the next sequence item from the request queue
        """
        return await self.export.get_next_item()

    def item_done(self, rsp=None):
        """Notify finish_item that the item is complete"""
        self.export.item_done(rsp)

    async def get_response(self, transaction_id=None):
        """
        Either get a response item with the given transaction_id,
        or get the first one in the queue.

        Removes the found transaction.

        If there is no transaction in the queue with transaction_id,
        block until it appears.
        """
        datum = await self.export.get_response(transaction_id)
        return datum


# The UVM sequencer is really just a holder for the
# seq_item_export that does all the work.


class uvm_sequencer(uvm_component):
    """
    The uvm_sequencer arbitrates between multiple sequences that want to send
    items to driver (connected to seq_export) It exposes put_req,
    get_next_item, get_response from the export.
    The sequence will use these to coordinate
    items with the sequencer.
    """

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.seq_item_export = uvm_seq_item_export("seq_item_export", self)
        self.seq_q = UVMQueue(0)

    async def run_phase(self):
        while True:
            next_item = await self.seq_q.get()
            next_item.start_condition.set()
            next_item.start_condition.clear()

    async def start_item(self, item):
        await self.seq_q.put(item)
        await item.start_condition.wait()

    async def finish_item(self, item):
        await self.seq_item_export.put_req(item)
        await item.finish_condition.wait()

    async def put_req(self, req):
        await self.seq_item_export.put_req(req)

    async def get_response(self, txn_id=None):
        datum = await self.seq_item_export.get_response(txn_id)
        return datum

    async def get_next_item(self):
        next_item = await self.seq_item_export.get_next_item()
        return next_item


class uvm_sequence(uvm_object):
    """
    The uvm_sequence creates a series of sequence
    items and feeds them to the sequencer
    using start_item() and finish_item(). It can
    also get back results with get_response()
    body() gets launched in a thread at start.
    """

    def __init__(self, name):
        super().__init__(name)
        self.sequencer = None
        self.running_item = None
        self.sequence_id = id(self)

    async def body(self):
        """
        This function gets launched in a thread when you run start()
        You generally override it in any extension.
        """

    async def start(self, seqr=None):
        if seqr is not None:
            assert (isinstance(seqr, uvm_sequencer)), \
                "Tried to start a sequence with a non-sequencer"
        self.sequencer = seqr
        await self.body()

    async def start_item(self, item):
        """
        Sends an item to the sequencer and waits to be notified
        when the item has been selected to be run.

        :param item: The sequence item to send to the driver.
        """
        if self.sequencer is None:
            raise error_classes.UVMSequenceError(
                "Tried start_item in a virtual "
                f"sequence {self.get_full_name()}")
        item.parent_sequence_id = self.sequence_id
        self.running_item = item
        await self.sequencer.start_item(item)

    async def finish_item(self, item):
        if self.sequencer is None:
            raise error_classes.UVMSequenceError(
                "Tried finish_item in virtual"
                f" sequence: {self.get_full_name()}")
        await self.sequencer.finish_item(item)

    async def get_response(self):
        if self.sequencer is None:
            raise error_classes.UVMSequenceError(
                "Tried to do get_response in a virtual "
                f"sequence: {self.get_full_name()}")
        datum = await self.sequencer.get_response(
            self.running_item.response_id)
        return datum



from pyuvm.s05_base_classes import uvm_object
import pyuvm.error_classes as error_classes
import cocotb

# 9.1
#
# This is a dramatically simplified version of UVM phasing. We don't have
# to deal with simulation time and we are not going to deal with a generalized
# phasing system.
#
# So this system simply traverses the common phases, calling the appropriate
# method in each component.
#
# Much of the work in the SV phasing code has to do with handling the passage
# of time.  There is no timewheel in Python, so all of that code can go
# away.
#
# Also, the generalized phasing system is rarely used and so that
# is left as an exercise for future developers.  Instead we have a simple
# topdown and bottom up traversal of calling methods in component
# classes based on the phase name.
#
# We're not doing schedules or domains. We're just creating a list of classes
# and traversing them in order. The order it dependent upon whether they
# are topdown or bottom up phases.


# 9.3.1.2 Class declaration
class uvm_phase(uvm_object):

    # Strips the "uvm_" from this class's name and uses the remainder
    # to get a function call out of the component and execute it.
    # 'uvm_run_phase' becomes 'run_phase' and is called as 'run_phase()'
    @classmethod
    def execute(cls, comp):
        """
        :param comp: The component whose turn it is to execute
        """
        method_name = cls.__name__[4:]
        try:
            method = getattr(comp, method_name)
        except AttributeError:
            raise error_classes.UVMBadPhase(
                f"{comp.get_name()} is missing {method_name} function")
        method()


class uvm_topdown_phase(uvm_phase):
    """
    Runs phases from the top down.
    """
    @classmethod
    def traverse(cls, comp):
        """
        Given a component, we traverse the component tree
        top to bottom calling the phase functions as we go

        :param comp: The component whose hierarchy will be traversed
        """
        cls.execute(comp)  # first we execute this node then its children
        for child in comp.get_children():
            cls.traverse(child)


class uvm_bottomup_phase(uvm_phase):
    """
    Runs the phases from bottom up.
    """
    @classmethod
    def traverse(cls, comp):
        for child in comp.get_children():
            cls.traverse(child)
        cls.execute(comp)


class uvm_threaded_execute_phase(uvm_phase):
    """
    This phase launches the phase function in a thread and
    returns the thread to the caller.  The caller can then
    join all the threads.
    """

    @classmethod
    def execute(cls, comp):
        phase_name = cls.__name__
        assert phase_name.startswith("uvm_"), \
            "We only support phases whose names start with uvm_"
        method_name = cls.__name__[4:]
        try:
            method = getattr(comp, method_name)
        except AttributeError:
            raise error_classes.UVMBadPhase(
                f"{comp.get_name()} is missing {method_name} function")
        cocotb.fork(method())


# 9.8 Predefined Phases
# 9.8.1 Common Phases
# The common phases are described in the order of their execution.
# 9.8.1.1
class uvm_build_phase(uvm_topdown_phase):
    ...


# 9.8.1.2
class uvm_connect_phase(uvm_bottomup_phase):
    ...


# 9.8.1.3
class uvm_end_of_elaboration_phase(uvm_topdown_phase):
    ...


# 9.8.1.4
class uvm_start_of_simulation_phase(uvm_topdown_phase):
    ...


# 9.8.1.5
class uvm_run_phase(uvm_threaded_execute_phase, uvm_bottomup_phase):
    ...


# 9.8.1.6
class uvm_extract_phase(uvm_topdown_phase):
    ...


# 9.8.1.7
class uvm_check_phase(uvm_topdown_phase):
    ...


# 9.8.1.8
class uvm_report_phase(uvm_topdown_phase):
    ...


# 9.8.1.9
class uvm_final_phase(uvm_topdown_phase):
    ...


# 9.8.2
# UVM run-time phases are left as an exercise for an enterprising soul
# I cannot imagine why anyone would implement this.
# One could add phases by simply extending uvm_topdown_phase
# or uvm_bottom_up phase with a new phase named 'uvm_my_phase' and adding
# the my_phase() method to a uvm component with setattr.

uvm_common_phases = [uvm_build_phase,
                     uvm_connect_phase,
                     uvm_end_of_elaboration_phase,
                     uvm_start_of_simulation_phase,
                     uvm_run_phase,
                     uvm_extract_phase,
                     uvm_check_phase,
                     uvm_report_phase,
                     uvm_final_phase]
